## Required imports ##
import kitsubot as k  # Imports all of the functions from the main module
perms = k.perms       # Allows @perms("x") decorator to be used
import config as c    # Allows access to users, other mods, and globals
######################

# Callbacks you do not intend to use can be safely removed, but if you
# do not have an init() you don't really have a module.  All other callbacks
# here are optional and provided for documentation only.  In fact, it is
# RECOMMENDED that you remove functions you don't use.

# Refer to config.py to see what is imported as the 'c' object.
# Also refer to the main bot file kitsubot.py to see what is imported by
# the 'k' object.

def help(h):
    """
    Help method, called when /help <module> is used.
    If more than one topic is needed, use the h parameter to direct
    the user:
        /help example topic -- will return 'topic' in h to this module
        if not specified, h is None.
    """
    return "\n".join([
        "Preferred method of setting up your help module",
        "Is to create a list right here."
    ])

def init(dp):
    """
    Init method, called when modules are loaded.  Please note
    that the module loader will not call your reload() method
    automatically on init.  You need to load your own config.
    Dispatcher is sent as 'dp'.
    Hook the commands and regexes you want to hook in this method,
    as well as any other initialization.
    """
    # Any hook from the 'telegram.dispatcher' class can be used here.
    # Responds to the command /test
    dp.addTelegramCommandHandler("test", cmd_test)
    # Responds to the word "test"
    dp.addTelegramRegexHandler("test",cmd_test)
    # If you have data you want to load from disk, you probably want to use
    reload()

def reload():
    """
    This function is called when the /reload command is issued and should
    either reload the configuration for the module or reset it.
    """
    pass

def commit():
    """
    This function is called at commit_interval in config.py, and whenever
    the /commit command is used.  You should save your local configuration
    to disk when this is called.
    """
    pass

def backup():
    """
    Called when /backup is used.  Save your files in an alternate location.
    And do whatever else you think you might need to do.
    """
    pass

def term():
    """
    This function is called just before the bot exits, but after commit is
    called.  If you have something to do before a shutdown, do it here.
    """
    pass
    
@perms("g")
def cmd_test(b,u):
    """
    As declared in the init module, this function will be called when
    the bot handler triggers it.
    b = the incoming bot object
    u = the specific update.
    """
    # The helper function here gets data in a more convenient form.
    # user = the user object from the update (basically, u.message.from_user)
    # chat = the chat id from the update (not the chat object)
    #        the chat object is at u.message.chat if you need it.
    # uid = the user id of the update
    # name = the username of the updating user
    # perm = the users permission level
    # cmd = the command issued
    # arg = the single string arguments after the command, or None if none were given
    # This is given to make getting useful information out of the update easier.
    user, chat, uid, name, perm, cmd, arg = k.getInfo(b,u)
    pass
