#!/usr/bin/env python
# -*- coding: utf-8 -*-

### For maximum compatibility, don't modify this file.  Thats what modules are for.
### But of course, who can stop you.
### This bot and the basic modules it is released with are licensed under the
### GNU Public License, version 3:
### http://www.gnu.org/licenses/gpl-3.0.en.html

### You will need at least these libraries for this to run:
###     pip3 install telegram-python-bot
###     pip3 install xmltodict
### This was written in python 3.4 but should work for any python 3.0 version
### and possibly with python2 and __future__
### Python 2 support was neither targetted or tested.

"""
Kitsubot!  The foxiest bot evar!
"""

from telegram.ext import Updater
from telegram.ext.dispatcher import run_async
from time import sleep
import logging
import sys, select, os
import config as c
import json
import traceback
import unicodedata
import xmltodict
import urllib.request
import urllib.error, socket

from enum import Enum


### The primary docs ###
docs = {
    "/start": "Start the bot, usually happens automagically.  Also display version.",
    "/ver": "List the bot version.",
    "/data": "List the data in datastores (in json).  Valid options are local, chat, sys (defaults to user)",
    "/help": "Display the docs for the commands"
}
docs_admin = {
    "/perms": "<user> <perm>: Sets a users permission level.  Only Root users can set a or r. [a]",
    "/commit": "Dump data to disk [o]",
    "/reload": "Reload data from disk [o]",
    "/backup": "Make a backup of the config [a]",
    "/disable": "<command> - disable this command for this channel [o]",
    "/enable": "<command> - enable this command for this channel. [o]",
    "/set": "<target> <prop> <value> - set a property - see /help props [a]",
    "/get": "<target> <prop> - get a property - see /help props [a]",
    "/stop": "Kill the bot [a]"
}

def aybXML(url):
    """
    Takes a URL, processes an XML response, returns either a dictionary
    (no error) or a string (error)
    """
    try:
        www = urllib.request.Request(url,None,headers={"User-agent": "Mozilla/5.0 (AmigaOS 3.1; m68k; idver:%s) OWB/19760721 %s/%s" % (c.botname, c.botbrand, c.version)})
        with urllib.request.urlopen(www,timeout=5) as f:
            return dict(xmltodict.parse(f.read().decode("utf-8")))
    except (urllib.error.HTTPError, urllib.error.URLError) as e:
        return e.msg
    except socket.timeout as e:
        return "5 second timeout"
    except Exception as e:
        traceback.print_exc()
        return "Not a AmigaDOS disk in drive DH0:"

def aybJSON(url):
    """
    Takes a URL, processes an JSON response, returns either a dictionary
    (no error) or a string (error)
    """
    try:
        www = urllib.request.Request(url,None,headers={"User-agent": "Mozilla/5.0 (AmigaOS 3.1; m68k; idver:%s) OWB/19760721 %s/%s" % (c.botname, c.botbrand, c.version)})
        with urllib.request.urlopen(www,timeout=5) as f:
            return dict(json.loads(f.read().decode("utf-8")))
    except (urllib.error.HTTPError, urllib.error.URLError) as e:
        return e.msg
    except socket.timeout as e:
        return "5 second timeout"
    except Exception as e:
        traceback.print_exc()
        return "Not a AmigaDOS disk in drive DH0:"

def listMod(l,s):
    """
    List modify function -- takes a space delimited string and adds members to a list
    unless prefixed by -, whereas it will then remove the list, and if the string is
    -*, it clears the list.
    """
    if s == "-*": return []
    s = s.split(" ")
    for j in s:
        if j.startswith("-"):
            try:
                del l[l.index(j[1:])]
            except: pass
        else:
            l.append(j)
    return l

def getUsername(u):
    """
    Returns a valid username from an update, and corrects strings to be a valid username.
    """
    if type(u) is not str: v = u.message.from_user.username
    else: v = u
    if v == "":
        try:    v = u.message.from_user.first_name+"_"+u.message.from_user.last_name
        except: pass
    if v == "":
        try:    v = u.message.from_user.id
        except: pass
    if v == "":
        v = None
    if v.startswith("@"): v = v[1:]
    return v

def umatch(u):
    """
    Try to match a username, first name, last name, or user id to a username
    This function requires that the user has been seen at least once.
    """
    if u.startswith("@"): u = u[1:]
    for x in c.user.keys():
        if u.lower() == x.lower(): return x
    for j in c.user:
        if getUser(j,"id","") == u: return j
        if getUser(j,"last_name","").lower() == u.lower(): return j
        if getUser(j,"first_name","").lower() == u.lower(): return j
    return None

def newUser(u):
    """This sets up a users initial datastore and gives them guest permissions."""
    if type(u) is str:
        if c.user[u].get("@perms",None) is None: c.user[u] = {"@perms": "g"}
    else:
        t = u.message.from_user
        u = t.username
        try:    c.user[u]
        except: c.user[u] = {}
        c.user[u]["first_name"]=t.first_name
        c.user[u]["last_name"]=t.last_name
        c.user[u]["id"]=t.id
        if c.user[u].get("@perms",None) is None: c.user[u]["@perms"] = "g"
    logger.info("New user created: %s" % u)

def setSys(k, v):
    """
    Write something to the system datastore.
    k = key to write to; v = value to write
    """
    try:
        if v is None:
            del c.sys[k]
        else:
            c.sys[k] = v;
    except:
        logger.warn("setSys: Failed [%s]=%s" % (k,repr(v)))

def getSys(k, default=None):
    """
    Read something from the system datastore, with a default to be specified.
    k = key to read; default=default value
    """
    try:    return c.sys[k]
    except: return default

def setChat(ch, k, v):
    """
    Write something to the chat datastore.
    ch = chat id; k = key to write to; v = value to write
    """
    ch=str(ch)
    try:
        c.chat[ch][k] = v
    except:
        try:
            c.chat[ch]={}
            if v is None:
                del c.chat[ch][k]
            else:
                c.chat[ch][k] = v
        except:    
            logger.warn("setChat: Failed [%s]=%s" % (k,repr(v)))

def getChat(ch, k, default=None):
    """
    Read something from the chat datastore, with a default to be specified.
    ch = chat id; k = key to read; default=default value
    """
    ch=str(ch)
    try:    return c.chat[ch][k]
    except:
        #print(ch,k)
        #traceback.print_exc()
        return default

def setUser(u, k, v):
    """
    Write something to the users datastore, creating it if necessary.
    u = username; k = key to write to; v = value to write
    """
    if c.user.get(u) is None: newUser(u)
    if v is None:
        del c.user[u][k]
    else:
        c.user[u][k] = v;

def getUser(u, k, default=None):
    """
    Read something from the datastore, with a default to be specified.
    u = username, k = key to write to; default=default value
    """
    try:    return c.user[u][k]
    except: return default
    
def getEnv(upd, u, k, default=None):
    """
    Read something from the environment.
    which is to say if u = ch, it writes to the the user database
    for a chat: c.chat[ch]["@"][username]
    else write to ones user db.
    upd = update instance; u = username; k = key to write to; default=default value
    """
    if upd.message.from_user.id == upd.message.chat_id:
        try:    return c.user[u][k]
        except: return default
    else:
        try:    return c.chat[str(upd.message.chat_id)]["@"][u][k]
        except:
            try:    return c.user[u][k]
            except: return default
            

def setEnv(upd, u, k, v):
    """
    Read something from the environment.
    which is to say if u = ch, it writes to the the user database
    for a chat: c.chat[ch]["@"][username]
    else write to ones user db.
    upd = update instance; u = username; k = key to write to; v = value
    """
    if upd.message.from_user.id == upd.message.chat_id: 
        if c.user.get(u) is None: newUser(upd)
        c.user[u][k] = v;
    else:
        ch = str(upd.message.chat_id)
        if c.chat.get(ch) is None: c.chat[ch] = {}
        if c.chat.get(ch).get("@") is None: c.chat[ch]["@"] = {}
        if c.chat.get(ch).get("@").get(u) is None: c.chat[ch]["@"][u] = {}
        if v is None:
            del c.chat[ch]["@"][u][k]
        else:
            c.chat[ch]["@"][u][k] = v

def getLoc(ch, u, k, default=None):
    """
    Read something users local chat data store.
    Does not seek up the environment.
    ch = chat id; u = username; k = key to write to; default=default value
    """
    try:    return c.chat[str(ch)]["@"][u][k]
    except: return default
            

def setLoc(ch, u, k, v):
    """
    Write to the users local chat data store.
    Does not seek up the environment.
    ch = chat id; u = username; k = key to write to; v = value
    """
    ch = str(ch)
    if c.chat.get(ch) is None: c.chat[ch] = {}
    if c.chat.get(ch).get("@") is None: c.chat[ch]["@"] = {}
    if c.chat.get(ch).get("@").get(u) is None: c.chat[ch]["@"][u] = {}
    if v is None:
        del c.chat[ch]["@"][u][k]
    else:
        c.chat[ch]["@"][u][k] = v    

def cmpPerm(uperm,fperm):
    if uperm not in c.permlevel: return None
    if fperm not in c.permlevel: return None
    return c.permlevel.index(uperm)-c.permlevel.index(fperm)

def getPerm(u, glob=False, lesser=False, override=None):
    """
    Return a users permissions.  Accepts username, or update object.
    This returns the greater permission of either global or chat,
    lesser: if True, return the lesser of the permissions.
    glob: if True, seek only the global permission.
    """
    ch = None
    if type(u) is not str: ch = u
    if override is None:
        u = getUsername(u)
    else:
        u = getUsername(override)
    if ch is not None:
        g = getUser(u, "@perms", "g")
        l = getLoc(ch.message.chat_id, u, "@perms", "g")
    else:
        g = getUser(u, "@perms", "g")
        l = g
    # If users are banned in either the current channel or globally
    # they are considered banned.  Its not overridable.
    if g == "b" or l == "b": return "b"
    # If looking for global permissions, return global permissions
    # immediately.
    if glob: return g
    if not lesser:
        if cmpPerm(g,l)>0:  return g
        else:               return l
    else:
        if cmpPerm(g,l)<0:  return g
        else:               return l

def testPerm(u,fperm, glob=False):
    """
    Tests if the user 'u' has the permission level 'fperm' or greater.  Returns a bool.
    """
    uperm = getPerm(u, glob=glob)
    # No access to functions without b perms if banned.
    if uperm == "r": return True
    if uperm == "b" and not fperm == "b":
        logger.warn("B&  (%s)" % getUsername(u))
        return False
    # Bail if permission level < function level, and user is not root
    if cmpPerm(uperm,fperm) < 0:
        logger.warn("HaX (%s) (%s<%s)" % (getUsername(u), uperm, fperm))
        return False
    return True

def getInfo(b,u):
    """
    Helper function to extract useful data from the bot and update objects.
    Recommended usage:
    user, chat, uid, name, perm, cmd, arg = k.getInfo(b,u)
    """
    user = u.message.from_user
    chat = u.message.chat_id
    uid = u.message.from_user.id
    name = getUsername(u)
    perm = getPerm(u)
    try:    cmd, arg = u.message.text.split(" ",1)[0], u.message.text.split(" ",1)[1]
    except: cmd, arg = u.message.text, None
    return user,chat,uid,name,perm,cmd,arg

def perms(fperm,local=False):
    """Decorator for users handlers.  @perms(".") means that only users
    with . level or greater will be able to use that handler.  Replace .
    with a legal permission.
    This also handles if a command is disabled for a chat.
    """
    def perms_decorator(func):
        def func_wrapper(b,u):
            disabled = getChat(u.message.chat_id,"@disabled",{})
            try:
                if disabled[u.message.text.replace("/","").split(" ")[0]]: return
                if disabled["*"]: return
            except: pass
            if testPerm(u,fperm,not local): return run_async(func(b,u))
            else: return
        return func_wrapper
    return perms_decorator

### Not documented -- don't modify            
def commit():
    with open("system.json",mode='w') as f: f.write(json.dumps(c.sys))
    with open("chats.json",mode='w') as f: f.write(json.dumps(c.chat))
    with open("users.json",mode='w') as f: f.write(json.dumps(c.user))
    for j in c.mods:
        try: j.commit();
        except: pass

### Not documented -- don't modify
def reload():
    try:
        with open("system.json") as f: c.sys = json.loads(f.read())
    except:
        c.sys = {}
    try:
        with open("chats.json") as f: c.chat = json.loads(f.read())
    except:
        c.chat = {}
    try:
        with open("users.json") as f: c.user = json.loads(f.read())
    except:
        c.user = {}
    # Enforce root 
    for j in c.root: setUser(j, "@perms", "r")
    commit()        
    for j in c.mods:
        try: j.reload();
        except: pass

### Its best not to mess with the main command handlers.  Make your own module. :) ###
### For that matter everything below this shouldn't be modified.                   ###
logging.basicConfig(filename=c.log,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        level=logging.INFO)
logger = logging.getLogger(__name__)
#stdoutLogger=logging.StreamHandler(sys.stdout)
#stdoutLogger.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
#logging.getLogger().addHandler(stdoutLogger)

last_chat_id = 0
suicide = False

def getDB(u, s, write=False):
    # Extract the prop value
    try:
        prop, val = s.rsplit("=",1)[0], s.rsplit("=",1)[1]
    except:
        prop, val = s, None
    # Get the value to its native type
    if val is not None:
        try:
            if "." in val:  val = float(val)
            else:           val = int(val)
        except:
            if val.lower() == "false": val = False
            if val.lower() == "true": val = True
            
    # Extract the prop key
    try:
        obj, prop = prop.split(":",1)[0], prop.split(":",1)[1]
    except:
        return None, None, None, val, -1
    try:
        if prop[0] in "@.": proptype = prop[0]
        else: proptype = ""
    except:
        return None, None, None, val, -2        

    # Extract the prop datastore
    try:
        obj, tu = obj.split("/")[0], obj.split("/")[1]
    except: tu = None

    # Determine the datastore
    if obj.lower() == "me":
        obj = u.message.from_user.username
    elif obj.lower() == "here":
        obj = u.message.chat_id
        if type(tu) is str:
            if tu.lower() == "me":
                tu = u.message.from_user.username
    try:
        obj = int(obj)
    except:
        if not obj == "@":
            obj = umatch(obj)
    

    ## Determine if we're allowed to 
    gperm = getPerm(u, glob=True)
    lperm = getPerm(u)
    print(repr((obj,tu,proptype,prop,val)))
    if not write:   # Read permissions
        # Disallow read of @ props by non-opers
        if proptype == "@" and cmpPerm(gperm,"o")<0: return None, None, None, None, 1     
        # Disallow read of users, except by themselves or opers
        if type(obj) is str:
            if not obj == "@":
                if obj != u.message.from_user.username and cmpPerm(gperm,"o")<0: return None, None, None, None, 2
        # Disallow read of local datastores except by those with oper rights in the channel, or the user.
        if type(obj) is int and tu is not None and(cmpPerm(lperm,"o")<0 or tu != u.message.from_user.username): return None, None, None, None, 3
    else:           # Write permissions
        # Disallow write to system by non-admins
        if obj == "@" and cmpPerm(gperm,"a")<0: return None, None, None, None, 4
        # Disallow write to users, except by themselves or non-admins
        if type(obj) is str and obj != u.message.from_user.username and cmpPerm(gperm,"a")<0: return None, None, None, None, 5
        # Disallow write to chats except by those with oper rights in the channel    
        if type(obj) is int and tu is None and cmpPerm(lperm,"o")<0: return None, None, None, None, 6
        # Disallow write to local datastore unless channel admin or the user:
        if type(obj) is int and tu is not None and cmpPerm(lperm,"a")<0 and tu != u.message.from_user.username: return None, None, None, None, 7
        # Disallow write to @ props by non-admins
        if proptype == "@" and type(obj) is str and cmpPerm(gperm,"a")<0: return None, None, None, None, 8
        if proptype == "@" and type(obj) is int and cmpPerm(lperm,"a")<0: return None, None, None, None, 9
        # Disallow write to . props by non-opers
        if proptype == "." and type(obj) is str and cmpPerm(gperm,"o")<0: return None, None, None, None, 10
        if proptype == "." and type(obj) is int and cmpPerm(lperm,"o")<0: return None, None, None, None, 11

    # Operation is permitted.
    return obj, tu, prop, val, 0

@perms("g")
def cmd_set(b, u):
    user, chat, uid, name, perm, cmd, arg = getInfo(b,u)
    obj, tu, prop, val, p = getDB(u, arg, write=True)
    try:
        if p > 0:
            b.sendMessage(chat,"Permission denied. (%d)" % p)
            return
        if p < 0:
            b.sendMessage(chat,"Kernel panic.  I can't believe they gave you a starship. (%d)" % p)
            return
        if obj == "@":
            setSys(prop,val)
            b.sendMessage(chat,"System property '%s' set: %s" % (prop, repr(val)))
        elif type(obj) is str:
            setUser(obj,prop,val)
            b.sendMessage(chat,"User @%s property '%s' set: %s" % (obj, prop, repr(val)))
        elif type(obj) is int and tu is None:
            setChat(obj,prop,val)
            b.sendMessage(chat,"Chat #%d property '%s' set: %s" % (obj, prop, repr(val)))
        elif type(obj) is int and tu is not None:
            setLoc(obj,tu,prop,val)
            b.sendMessage(chat,"Local #%d/%s property '%s' set: %s" % (obj, tu, prop, repr(val)))
    except:
        b.sendMessage(chat,"Something went terribly wrong:\n"+repr(traceback.format_exc()))

@perms("g")
def cmd_get(b, u):
    user, chat, uid, name, perm, cmd, arg = getInfo(b,u)
    obj, tu, prop, val, p= getDB(u, arg, write=False)
    try:
        if p > 0:
            b.sendMessage(chat,"Permission denied. (%d)" % p)
            return
        if p < 0:
            b.sendMessage(chat,"Kernel panic.  I can't believe they gave you a starship. (%d)" % p)
            return
        if obj == "@":
            val = getSys(prop,None)
            b.sendMessage(chat,"@:%s = %s" % (prop, repr(val)))
        elif type(obj) is str:
            val = getUser(obj,prop,None)
            b.sendMessage(chat,"@%s:%s = %s" % (obj, prop, repr(val)))
        elif type(obj) is int and tu is None:
            val = getChat(obj,prop,None)
            b.sendMessage(chat,"#%d:%s = %s" % (obj, prop, repr(val)))
        elif type(obj) is int and tu is not None:
            val = getLoc(obj,tu,prop,None)
            b.sendMessage(chat,"#%d/%s:%s = %s" % (obj, tu, prop, repr(val)))
    except:
        b.sendMessage(chat,"Something went terribly wrong:\n"+repr(traceback.format_exc()))
                    
@perms("g")
def cmd_start(b, u):
    """What does the fox say?"""
    user, chat, uid, name, perm, cmd, arg = getInfo(b,u)
    topics = []
    for x in os.listdir("help/"):
        topics.append(x.replace(".txt",""))
    b.sendMessage(uid, "\n".join([
        "%s v%s (C) 2016 Alynna Trypnotk, License GPLv3" % (c.botbrand, c.version),
        "Type /help for help",
        "Modules loaded:",
        "  :  ".join(c.mods.keys()),
        "Topics available:",
        "  :  ".join(topics),
        "Type /help <module/topic> for more help."]))
    #logger.info(repr((user, chat, uid, name, perm, cmd, arg)))
    if cmpPerm(perm, "u") < 0:
        setUser(name, "@perms", "u");
        b.sendMessage(chat,"You are now registered at this shrine.")
    else:
        b.sendMessage(chat,"You are already registered at this shrine.")

@perms("a")
def cmd_stop(b, u):
    """ Stop the Bot """
    user, chat, uid, name, perms, cmd, arg = getInfo(b,u)
    global suicide, dp
    suicide = True
    if arg is not None:
        b.sendMessage(chat, 'Shutting down without save.')
        return
    b.sendMessage(chat, "Saving and shutting down.")
    commit()
    for j in c.mods: 
        try:    c.mods[j].commit(dp)
        except: pass

@perms("o")
def cmd_reload(b, u):
    """ reload the bot configuration """
    user, chat, uid, name, perms, cmd, arg = getInfo(b,u)
    reload()
    logger.info("Reloading {botbrand} configuration" % {"botbrand":c.botbrand})
    for j in c.mods: 
        try:
            c.mods[j].reload()
            logger.info("Reloading configuration "+j)
        except: pass
    b.sendMessage(chat, "Reloaded bot configuration.")

@perms("o")
def cmd_commit(b, u):
    """ reload the bot configuration """
    user, chat, uid, name, perms, cmd, arg = getInfo(b,u)
    commit()
    logger.info("Saving {botbrand} configuration" % {"botbrand":c.botbrand})
    for j in c.mods: 
        try:
            c.mods[j].commit()
            logger.info("Saving configuration "+j)
        except: pass
    b.sendMessage(chat, "Saved bot configuration.")

@perms("a")
def cmd_backup(b, u):
    """ This saves to a backup, and requests modules to do so too. """
    user, chat, uid, name, perms, cmd, arg = getInfo(b,u)
    logger.info("Making a backup of bot config ...")
    with open("system.json.bak",mode='w') as f: f.write(json.dumps(c.sys))
    with open("chat.json.bak",mode='w') as f: f.write(json.dumps(c.chat))
    with open("users.json.bak",mode='w') as f: f.write(json.dumps(c.user))
    logger.info("Requesting modules to backup ...")
    for j in c.mods: 
        try:
            c.mods[j].backup()
            logger.info(j+" Backing up ..")
        except: pass
    b.sendMessage(chat, "Config backup complete.")
    
def unknown_command(b, u):
    # This was annoying.  Active but doesn't do anything.
    pass
    # b.sendMessage(update.message.chat_id, text='Command not recognized!')

@perms("g")
def cmd_perms(b,u):
    user, chat, uid, name, perms, cmd, arg = getInfo(b,u)
    if arg is None:
        b.sendMessage(chat,"Your access levels: Global [%s]; %s [%s]; Effective [%s]" %
                      (getPerm(u, glob=True), u.message.chat.title, getLoc(chat, name, "@perms","g"), getPerm(u)))
        return
    if not cmpPerm(perms, "o") >= 0:
        b.sendMessage(uid,"Permission denied.")
        return
    try:
        target, perm = arg.split(" ")[0], arg.split(" ")[1]
        target = umatch(target)
    except:
        target, perm = umatch(arg), None
    if perm is None:
        b.sendMessage(chat,"@%s access levels: Global [%s]; %s [%s]; Effective [%s]" %
                      (target, getPerm(u, glob=True, override=target),u.message.chat.title,
                       getLoc(chat, target, "@perms","g"), getPerm(u, override=target)))
        return        
    if not perm in ".bgumoa":
        b.sendMessage(chat,"Invalid permission.  Seek /help perms.")
        return
    if not perm == ".":
        if cmpPerm(perms,perm) < 1:
            logger.warn("[HaX] /perms: %s->%s : %s->%s" % (name, target, perms, perm))
            b.sendMessage(chat,"You are NOT permitted to set that permission level.  This incident has been reported.")
            return
        if (chat == uid or cmd[1] == "g"):
            if cmpPerm(getPerm(name,glob=True),"a") >= 0:
                setUser(target, "@perms", perm)
                b.sendMessage(chat, "Global permission [%s] for user [%s] set." % (perm, target))
            else:
                logger.warn("[HaX] /perms: %s->%s : %s->%s (global)" % (name, target, perms, perm))
                b.sendMessage(chat,"You are NOT permitted to set global perms at your permission level.  This incident has been reported.")
                return
        else:
            setLoc(chat, target, "@perms", perm)
            b.sendMessage(chat, "Local permission [%s] for user '%s' in chat '%s' set." % (perm, target, u.message.chat.title))
    else:
        if cmpPerm(getPerm(name, glob=True),"a") < 0:
            logger.warn("[HaX] /perms: %s->%s : %s->%s" % (name, target, perms, perm))
            b.sendMessage(chat,"You are NOT permitted to delete users at your permission level.  This incident has been reported.")
            return
        try:
            del c.user[target]
            b.sendMessage(chat,"User deleted.")
        except:
            b.sendMessage(chat,"User not found.")
        
            
@perms("g")
def cmd_data(b,u):
    user, chat, uid, uname, perm, cmd, arg = getInfo(b,u)
    if arg is None:
        b.sendMessage(chat,"Your data\n"+repr(c.user[uname]))
        return
    elif arg.lower() == "chat":
        temp = c.chat[str(chat)].copy()
        del temp["@"]
        if testPerm(uname,"o"): b.sendMessage(chat,"Chat data\n"+repr(temp))
        else:                   b.sendMessage(chat,"Permission denied.  This incident has been reported.")
    elif arg.lower() == "sys":
        if testPerm(uname,"a",glob=True): b.sendMessage(chat,"System data\n"+repr(c.sys))
        else:                   b.sendMessage(chat,"Permission denied.  This incident has been reported.")
    elif arg.lower() == "local":
        try:    b.sendMessage(chat,"Your local chat data\n"+repr(c.chat[str(chat)]["@"][uname]))
        except: b.sendMessage(chat,"You have no local chat data.")
        return
    else:
        if testPerm(uname,"o", glob=True):
            try:    b.sendMessage(chat,"User data\n"+repr(c.user[umatch(arg)]))
            except: b.sendMessage(chat,"User not known: "+arg)
        else:   b.sendMessage(chat,"Permission denied.  This incident has been reported.")
            
@perms("o", local=True)
def cmd_enable(b,u):
    user, chat, uid, uname, perm, cmd, arg = getInfo(b,u)
    disabled = getChat(chat, "@disabled", {})
    if arg is None:
        b.sendMessage(chat,"Disabled commands for %s:\n%s" % (u.message.chat.title, "; ".join(disabled.keys())))
        return
    try:
        del disabled[arg]
    except:
        b.sendMessage(chat,"Command currently not disabled: "+arg)
        return
    setChat(chat, "@disabled", disabled)
    b.sendMessage(chat,"Command enabled: "+arg)

@perms("o", local=True)
def cmd_disable(b,u):
    user, chat, uid, uname, perm, cmd, arg = getInfo(b,u)
    disabled = getChat(chat, "@disabled", {})
    if arg is None:
        b.sendMessage(chat,"Disabled commands for %s:\n%s" % (u.message.chat.title, "; ".join(disabled.keys())))
        return
    disabled[arg] = True
    setChat(chat, "@disabled", disabled)
    b.sendMessage(chat,"Command disabled: "+arg)
    
@perms("g")
def cmd_help(b, u):
    """ Self generating docs. """
    user, chat, uid, uname, perm, cmd, arg = getInfo(b,u)
    helpmsg = "%s!\n`User commands:\n" % c.botbrand
    if arg is None:
        for x in docs:
            helpmsg += "{0:<10} {1}\n".format(x,docs[x])
        if testPerm(uname,"m"):
            helpmsg += "\nAdmin Commands\n"
            for x in docs_admin:
                helpmsg += "{0:<10} {1}\n".format(x,docs_admin[x])
        b.sendMessage(uid, helpmsg+"`", parse_mode="Markdown")
        topics = []
        for x in os.listdir("help/"):
            topics.append(x.replace(".txt",""))
        b.sendMessage(uid, "`Modules loaded:\n"+"  :  ".join(c.mods.keys())+"\n"+
                           "Topics available:\n"+"  :  ".join(topics)+"\nType /help <module/topic> for more help.`",parse_mode="Markdown")
    else:
        try:
            mod, topic = arg.split(" ")[0],arg.split(" ")[1]
        except:
            mod, topic = arg, None
        try:
            b.sendMessage(uid, "`"+c.mods[mod].help(topic)+"`")
        except:
            try:
                with open("help/"+arg+".txt") as f:
                    b.sendMessage(uid, "`%s`" % f.read(),parse_mode="Markdown")
            except:
                b.sendMessage(uid, "No help is available for that command.\nContact the writer of this module.")
    if not chat == uid:
        b.sendMessage(chat, "Help can only be requested in a private chat.  Click @%s and use /help." % c.botname)

@perms("o",local=True)
def cmd_say(b,u):
    user, chat, uid, uname, perm, cmd, arg = getInfo(b,u)
    target = getUser(uname,"@last_say_target",chat)
    try:
        if arg[0] in "@#?":
            tgt, arg, ctype = arg.split(" ",1)[0][1:], arg.split(" ",1)[1], arg[0]
        else:
            tgt, arg, ctype = target, arg, ""
    except:
        if arg[0] in "@#?":
            tgt, arg, ctype = target, arg[1:], arg[0]
        else:
            tgt, arg, ctype = target, arg, ""
    if ctype == "?":
        try:
            target = getUser(umatch(arg),"id",0)
            b.sendMessage(uid,"@%s's ID: %d; chat ID: %d" % (arg,target,chat))
            setUser(uname,"@last_say_target",target)    
        except:
            b.sendMessage(uid,"@%s's ID: %d; chat ID: %d" % (uname,getUser(uname,"id",0),chat))
        return
    elif ctype == "@":
        if testPerm(u,"o",glob=True)<0:
            b.sendMessage(uid,"Only global operators can privately message users through the bot.")
            return
        else:
            try:
                target = getUser(umatch(tgt),"id",0)
                if type(target) is str: target = int(target)
            except:
                b.sendMessage(uid,"Invalid target.")
                return
    elif ctype == "#":
        if not testPerm(u,"o",glob=True) and cmpPerm(getLoc(chat,uname,"@perms","g"),"o")<0:
            b.sendMessage(uid,"Only global operators can message channels through the bot.")
            return
        else:
            try:
                target = int(tgt)
            except:
                b.sendMessage(uid,"Invalid target.")
                return
    if target:
        b.sendMessage(target, arg)
    else:
        b.sendMessage(uid, "User not found.")        
    setUser(uname,"@last_say_target",target)

@perms("g")
def cmd_ver(b,u):
    b.sendMessage(u.message.chat_id,"%s v%s (C) 2016 Alynna Trypnotk, License GPLv3" % (c.botname, c.version))

# Called unconditionally
@run_async
def any_message(b, u, **kw):
    """ Base handler for incoming messages."""
    # Save last chat_id to use in reply handler
    global last_chat_id
    last_chat_id = u.message.chat_id
    # Register user:
    try:
        c.user[u.message.from_user.username]["id"]
    except:
        newUser(u)

    # Always register chat to track title changes.
    setChat(u.message.chat_id, "title", u.message.chat.title)
        
    # Log it.
    try:
        logger.info("(%s[%d] @ %s[%d] [%s]):\n    %s" %
            (u.message.from_user.username,u.message.from_user.id,
             u.message.chat.title, u.message.chat_id,
             getPerm(u),
             str(u.message.text.encode('utf-8','replace').decode('ascii','replace'))))
    except:
        logger.info("(%s[%d] @ %s[%d] [%s]):\n    %s" %
            (u.message.from_user.username,u.message.from_user.id,
             u.message.chat.title, u.message.chat_id,
             getPerm(u),
             str(u.message.text.encode('utf-8','replace'))))

# Called unconditionally
#@run_async
#def message(b, u, **kwargs):
#    """
#    Example for an asynchronous handler. It's not guaranteed that replies will
#    be in order when using @run_async. Also, you have to include **kwargs in
#    your parameter list. The kwargs contain all optional parameters that are
#    """
#    pass

def error(b, u, error):
    """ Print error to console """
    logger.warn('Update %s caused error %s' % (u, error))

### Main Screen Turn On -- Best not to modify ###
def main():
    # Create the EventHandler and pass it your bot's token.
    updater = Updater(c.token, workers=64)
    c.updater = updater
    global suicide, docs, docs_admin

    # Get the dispatcher to register handlers
    dp = updater.dispatcher
    c.dp = dp

    # This is how we add handlers for Telegram messages
    dp.addTelegramCommandHandler("start", cmd_start)
    dp.addTelegramCommandHandler("perms", cmd_perms)
    dp.addTelegramCommandHandler("gperms", cmd_perms)
    dp.addTelegramCommandHandler("data", cmd_data)
    dp.addTelegramCommandHandler("help", cmd_help)
    dp.addTelegramCommandHandler("stop", cmd_stop)
    dp.addTelegramCommandHandler("reload", cmd_reload)
    dp.addTelegramCommandHandler("commit", cmd_commit)
    dp.addTelegramCommandHandler("backup", cmd_backup)
    dp.addTelegramCommandHandler("disable", cmd_disable)
    dp.addTelegramCommandHandler("enable", cmd_enable)
    dp.addTelegramCommandHandler("set", cmd_set)
    dp.addTelegramCommandHandler("get", cmd_get)
    dp.addTelegramCommandHandler("say", cmd_say)
    dp.addTelegramCommandHandler("ver", cmd_ver)
    #dp.addUnknownTelegramCommandHandler(unknown_command)

    # Message handlers only receive updates that don't contain commands
    #dp.addTelegramMessageHandler(message)

    # Regex handlers will receive all updates on which their regex matches
    dp.addTelegramRegexHandler('.*', any_message)

    # All TelegramErrors are caught for you and delivered to the error
    # handler(s). Other types of Errors are not caught.
    dp.addErrorHandler(error)

    # Load bot config
    reload()

    skip = ("kitsubot.py", "example.py", "config.py")
    # Load modules!  Copy example.py as needed.
    # Read example.py for more info.
    x = os.listdir()
    # Load all mods
    for j in x:
        if not j in skip:
            if not j.endswith(".py"): continue
            c.mods[j.replace(".py","")] = None
    # Initialize all mods, and populate help
    for j in c.mods:
        logger.info("Loading module %s ..." % j)
        c.mods[j] = __import__(j.replace(".py",""), globals=[updater,dp,suicide]);
        c.mods[j].init(dp)
            
    # Start the Bot and store the update Queue, so we can insert updates
    update_queue = updater.start_polling(poll_interval=0.1, timeout=10)

    t = 0
    while True:
        sleep(1)
        t=(t+1) % c.commit_interval
        if suicide:
            updater.stop()
            for j in c.mods: 
                try:    c.mods[j].term()
                except: pass
            break
        if not t:
            commit()
            for j in c.mods: 
                try:    c.mods[j].commit()
                except: pass
    sys.exit(0)

if __name__ == '__main__':
    main()
    
