## Required imports ##
import kitsubot as k  # Imports all of the functions from the main module
perms = k.perms       # Allows @perms("x") decorator to be used
import config as c    # Allows access to users, other mods, and globals
######################

# Callbacks you do not intend to use can be safely removed, but if you
# do not have an init() you don't really have a module.  All other callbacks
# here are optional and provided for documentation only.  In fact, it is
# RECOMMENDED that you remove functions you don't use.

# Refer to config.py to see what is imported as the 'c' object.
# Also refer to the main bot file kitsubot.py to see what is imported by
# the 'k' object.

def help(h):
    """Help method, called when /help <module> is used."""
    if h is None:
        return "\n".join([
            "UserProfile module - Keeps some local profile information you can display",
            "/name <name>        : Set your userprofile name",
            "/gender <gender>    : Set your userprofile gender",
            "/species <species>  : Set your userprofile species",
            "/desc <descr>       : Set a description up to 255 characters",
            "/rating <r>         : How you like your RP (G, PG, PG13, R, X, XXX)",
            "/pic <url>          : An address to a picture that will be displayed",
            "/look [user]        : Display yours or anothers profile in the chat.",
            " ",
            "/help userprofile kinks : Learn about the kink setting commands!",
            " ",
            "Important note!  If you set these in a channel, it will only change for that channel.",
            "If you change it in a private chat to the bot, it will change the global default.",
            "When being read, it will use the channel set value first, then the default."
        ])
    elif h.lower() == 'kinks':
        return "\n".join([
            "UserProfile module - Kinks commands",
            "/kfave [list]   : A list of your favorite kinks!",
            "/kyes [list]    : A list of things you would always do.",
            "/kmaybe [list]  : A list of things you might do sometimes..",
            "/kno [list]     : A list of things you will never do.",
            "Kink [list]s use the list protocol to be set.",
            "Use /help lists to learn more!"
            ])

def init(dp):
    dp.addTelegramCommandHandler("name", cmd_set)
    dp.addTelegramCommandHandler("species", cmd_set)
    dp.addTelegramCommandHandler("gender", cmd_set)
    dp.addTelegramCommandHandler("desc", cmd_set)
    dp.addTelegramCommandHandler("rating", cmd_set)
    dp.addTelegramCommandHandler("pic", cmd_set)
    dp.addTelegramCommandHandler("look",cmd_look)
    dp.addTelegramCommandHandler("kinks",cmd_kinks)
    dp.addTelegramCommandHandler("kfave",cmd_kset)
    dp.addTelegramCommandHandler("kyes",cmd_kset)
    dp.addTelegramCommandHandler("kmaybe",cmd_kset)
    dp.addTelegramCommandHandler("kno",cmd_kset)

@perms("g")
def cmd_look(b,u):
    """
    As declared in the init module, this function will be called when
    the bot handler triggers it.
    b = the incoming bot object
    u = the specific update.
    """
    # The helper function here gets data in a more convenient form.
    # user = the user object from the update (basically, u.message.from_user)
    # chat = the chat id from the update (not the chat object)
    #        the chat object is at u.message.chat if you need it.
    # uid = the user id of the update
    # name = the username of the updating user
    # perm = the users permission level
    # cmd = the command issued
    # arg = the single string arguments after the command, or None if none were given
    # This is given to make getting useful information out of the update easier.
    user, chat, uid, name, perm, cmd, arg = k.getInfo(b,u)
    if arg is not None:
        name = k.umatch(arg)
    if name is None:
        b.sendMessage(chat,"Couldn't find that user, I tried..")
        return
    x = k.getEnv(u, name, "pic", "")+"\n"
    x += k.getEnv(u, name, "name", "(Name unset)")+"\n"
    x += "(Rating "+k.getEnv(u, name, "rating", "unset")+"): "+k.getEnv(u, name, "gender", "(Unknown gender)")+" "+k.getEnv(u, name, "species", "(Unknown species)")+"\n"
    x += k.getEnv(u, name, "desc", "")[:255]
    b.sendMessage(chat,x)

@perms("g")
def cmd_set(b,u):
    user, chat, uid, name, perm, cmd, arg = k.getInfo(b,u)
    cmd = cmd[1:]
    if not arg:
        k.setEnv(u, name, cmd, None)
        if (chat == uid):   b.sendMessage(chat,"Global default property '%s' cleared." % cmd)
        else:               b.sendMessage(chat,"Local channel property '%s' in '%s' cleared." % (cmd, u.message.chat.title))
    else:
        k.setEnv(u, name, cmd, arg)
        if (chat == uid):   b.sendMessage(chat,"Global default property '%s' set: %s" % (cmd,arg))
        else:               b.sendMessage(chat,"Local channel property '%s' in '%s' set: %s" % (cmd, u.message.chat.title, arg))

@perms("g")
def cmd_kset(b,u):
    user, chat, uid, name, perm, cmd, arg = k.getInfo(b,u)
    cmd = cmd[2:]
    if chat == uid: kinks = k.getUser(name, "kinks", {"fave": [], "yes": [], "maybe": [], "no": []})
    else:           kinks = k.getLoc(chat, name, "kinks", {"fave": [], "yes": [], "maybe": [], "no": []})
    if not arg:
        b.sendMessage(chat, "@%s %s kinks: %s" % (name, cmd.upper(), repr(kinks[cmd])))
    else:
        try:
            kinks[cmd] = k.listMod(kinks[cmd],arg)
        except:
            pass
        if (chat == uid):
            try:
                k.setUser(name, "kinks", kinks)
                b.sendMessage(chat,"Global default %s kinks set: %s" % (cmd.upper(),repr(kinks[cmd])))
            except:
                b.sendMessage(chat,"Global default %s kinks empty.")
        else:
            try:
                k.setLoc(chat, name, "kinks", kinks)
                b.sendMessage(chat, "@%s has altered their %s kinks.  Pray they do not alter them further: %s" % (name, cmd.upper(), repr(kinks[cmd])))
            except:
                b.sendMessage(chat, "Kinks empty, reverting to global defaults.")

@perms("g")
def cmd_kinks(b,u):
    user, chat, uid, name, perm, cmd, arg = k.getInfo(b,u)
    if arg is None: arg = name
    else:           arg = k.umatch(arg)
    mykinks = {"fave": [], "yes": [], "maybe": [], "no": []}
    x = k.getUser(arg, "kinks", {})
    for j in x: mykinks[j] += x[j]
    x = k.getLoc(chat, arg, "kinks", {})
    for j in x: mykinks[j] += x[j]
    
    x = "@%s's kinks:\n" % arg
    try:    x += "FAVE:  "+"; ".join(mykinks["fave"])+"\n"
    except: x += "FAVE:  (unset)"+"\n"
    try:    x += "YES:   "+"; ".join(mykinks["yes"])+"\n"
    except: x += "YES:   (unset)"+"\n"
    try:    x += "MAYBE: "+"; ".join(mykinks["maybe"])+"\n"
    except: x += "MAYBE: (unset)"+"\n"
    try:    x += "NO:    "+"; ".join(mykinks["no"])+"\n"
    except: x += "NO:    (unset)"+"\n"
    b.sendMessage(chat,x)
    
